@@ -1,17 +1,29 @@
 # -*- coding: utf-8 -*-
 
 import re
+import os
+import random
 from copy import copy
-
-try:
-    import pycld2
-except ImportError:
-    pycld2 = None
+import pycld2
 
 PADDINGS = ['30px', '60px', '80px']
 COLOR = '#004161'
 
 
+"""Utilies"""
+
+def detect_language(text):
+    raw_text = re.sub(r'(<[^>]*>)', r' ', text)
+    raw_text = re.sub(r'[^a-zA-Z\säöüß]', r'', raw_text)
+    raw_text = re.sub(r'( \w )', r' ', raw_text)
+    raw_text = ' '.join(random.sample(raw_text.split(),100))   # pick randomly 100 words
+    reliable, index, top_3_choices = pycld2.detect(raw_text, bestEffort=False)
+    if not reliable:
+        reliable, index, top_3_choices = pycld2.detect(raw_text, bestEffort=True)
+
+    return top_3_choices[0][1]
+
+
 """HTML mappings"""
 
 def style_mappings(text, title=None):
@@ -25,8 +37,16 @@ def style_mappings(text, title=None):
     text = add_copyright(text)
     text = format_bible_verses(text, title)
     text = format_lists(text)
-    text = format_language(text)
     assertion_test(text, text_raw, title)
+
+    lang = detect_language(text_raw)
+    text = format_language(text, lang)
+
+    if lang == 'de':
+        bible_check(text, title)
+
+    text = final_cut(text)
+    #text = format_ascii(text)  # Do we need this?
     return text
 
 
@@ -40,6 +60,7 @@ def standardize(text):
         text = text.replace('</h3><', '</h3>\n<')
         text = text.replace('</p><', '</p>\n<')
         text = text.replace('</li><', '</li>\n<')
+        text = text.replace('</table><', '</table>\n<')
 
         text = text.replace('<p> ', '<p>')
         text = text.replace('. </p>', '.</p>')
@@ -61,10 +82,13 @@ def standardize(text):
         text = text.replace('</ul></li>', '')
 
         text = text.replace('  ', ' ')
+        text = text.replace('\t', '')
 
-        # not compatible with py27 due to non-ascii characters
-        #text = text.replace(' ', ' ')
-        #text = text.replace('­', '')
+        try:  # not compatible with py27 due to non-ascii characters
+            text = text.replace(' ', ' ')
+            text = text.replace('­', '')
+        except:
+            pass
     return text
 
 
@@ -85,7 +109,7 @@ def regexp_style_mappings(text):
 
         # header4 = starts with [0-9] item
         text = re.sub(r'(<p>)(<b>)(\d{1,3})(\. )(\()(\d{1,3})(.*)(</b>)(</p>)', r'<h4>\3\4\5Vers \6\7</h4>', text)
-        text = re.sub(r'(<p>)(|<b>)(\d{1,3})(\. )(\()(\d{1,3})(.*)(</p>)', r'<h4>\3\4\5Vers \6\7</h4>', text)
+        text = re.sub(r'(<p>)(|<b>|<em>)(\d{1,3})(\. )(\()(\d{1,3})(.*)(</p>)', r'<h4>\3\4\5Vers \6\7</h4>', text)
 
         # header4 -> add 'Vers'
         text = re.sub(r'(<h4>\d{1,3}\. )(\(Vers )(\d{1,3})( )(\w)', r'\1\2\3\5', text)
@@ -105,19 +129,31 @@ def regexp_style_mappings(text):
         verse_style = '<p style="font-weight: bold; color:{};">'.format(COLOR)
         text = re.sub(r'(<h4>)(.*)(</h4>)([\r\n]+)(<p>)(.*)(</p>)', r'\1\2\3\4{} \6\7'.format(verse_style), text)
 
-        # ordered lists -> add paddings to [a-n], [iv]
+        # try to handle word-specific <ol>
+        vals = ['v', 'iv', 'iii', 'ii', 'i']
+        for i in range(len(vals)-1):
+            text = re.sub(r'(<ol><li>)(.*)(</li>[\r\n]</ol>)(<p>)({}\.)'.format(vals[i]), r'<p>{}.\2</p>\n\4\5'.format(vals[i+1]), text)
+
+        vals = ['g', 'f', 'e', 'd', 'c', 'b', 'a']
+        for i in range(len(vals) - 1):
+            text = re.sub(r'(<ol><li>)(.*)(</li>[\r\n]</ol>)(<p>)({}\.)'.format(vals[i]), r'<p>{}.\2</p>\n\4\5'.format(vals[i + 1]), text)
+
+        # ordered lists -> add paddings to [a-n], [ivx]
         pad0 = r'<p style="padding-left: {};">'.format(PADDINGS[0])
         pad1 = r'<p style="padding-left: {};">'.format(PADDINGS[1])
 
         text = re.sub(r'(<p>)(|<b>)([a-hj-u])(\.)', r'{}\2\3\4'.format(pad0), text)
         text = re.sub(r'(<p>)(|<b>)([a-hj-u])( )', r'{}\2\3.\4'.format(pad0), text)
-        text = re.sub(r'({}h\. )(.*)(</p>[\r\n])(<p>)(i\. <b>)'.format(pad1), r'\1\2\3{}\5'.format(pad0), text)
-        
-        text = re.sub(r'(<p>)(|<b>)([iv]+)(\.)', r'{}\2\3\4'.format(pad1), text)
-        text = re.sub(r'(<p>)(|<b>)([iv]+)( )', r'{}\2\3.\4'.format(pad1), text)
+        text = re.sub(r'({}h\. )(.*)(</p>[\r\n])(<p>)(i\. )(.*)(</p>[\r\n]<p>ii\.)'.format(pad0), r'\1\2\3{}\5\6\7'.format(pad1), text)
+        text = re.sub(r'({}h\. )(.*)(</p>[\r\n])(<p>)(i\. <b>)'.format(pad0), r'\1\2\3{}\5'.format(pad0), text)
+        text = re.sub(r'({}h\. )(.*)(</p>[\r\n])({})(i\. )(.*)(</p>[\r\n])({})(i\. )(.*)(</p>[\r\n])({})(i\. )'.format(pad0, pad1, pad1, pad1), r'\1\2\3{}\5\6\7{}\9\10\11{}\13'.format(pad1, pad0, pad1), text)
+        text = re.sub(r'({}[ivx]+\. )(.*)(</p>[\r\n])({})(i\. )(.*)(</p>[\r\n])({})(i\. )(.*)(</p>[\r\n])({})(ii\. )'.format(pad1, pad1, pad1, pad1), r'\1\2\3{}\5\6\7\8\9\10\11\12\13'.format(pad0), text)
+
+        text = re.sub(r'(<p>)(|<b>)([ivx]+)(\.)', r'{}\2\3\4'.format(pad1), text)
+        text = re.sub(r'(<p>)(|<b>)([ivx]+)( )', r'{}\2\3.\4'.format(pad1), text)
 
         # fix list item i. if it belongs to [a-n] hierarchy
-        text = re.sub(r'({}[iv]+\. )(.*)(</p>[\r\n])({})(i\. <b>)'.format(pad1, pad1), r'\1\2\3{}\5'.format(pad0), text)
+        text = re.sub(r'({}[ivx]+\. )(.*)(</p>[\r\n])({})(i\. <b>)'.format(pad1, pad1), r'\1\2\3{}\5'.format(pad0), text)
 
         # fix -> replace ii. with i. if it comes right after [a-z]\.
         text = re.sub(r'({}[a-hj-u]\. )(.*)(</p>\n{})(ii. )'.format(pad0, pad1), r'\1\2\3i. ', text)
@@ -153,6 +189,10 @@ def regexp_style_mappings(text):
         # add space before '('
         text = re.sub(r'([^\s])(\()', r'\1 \2', text)
 
+    # header 4 in preface
+    text = re.sub(r'(<p>)(|<b>|<em>)(\d{1,3}\. )(.*)(</p>)', r'<h4>\3\4</h4>', text)
+    text = re.sub(r'\.</h', '</h', text)
+
     return text
 
 
@@ -207,10 +247,15 @@ def format_lists(text):
     text = re.sub(r'({}a. )'.format(pad0), r'<ol>\n\1', text)
     text = re.sub(r'({}i. )'.format(pad1), r'  <ol class="i">\n\1', text)
     text = re.sub(r'({})([a-z])(\. )(.*)(</p>)'.format(pad0), r' <li>\4</li>', text)
-    text = re.sub(r'({})([iv]+)(\. )(.*)(</p>)'.format(pad1), r'   <li>\4</li>', text)
+    text = re.sub(r'({})([ivx]+)(\. )(.*)(</p>)'.format(pad1), r'   <li>\4</li>', text)
     text = re.sub(r'({} )(.*)(</p>)'.format(pad2), r'     <li>\2</li>', text)
 
+    # move table up if part of list
+    for space in [' ', '   ', '     ']:
+        text = re.sub(r'({}<li>.*</li>[\r\n])(<table>)([\s\S]+)(</table>)([\r\n]{}<li>)'.format(space, space), r'\1{}<table>\3</table>\5'.format(space), text)
+
     # list -> closings detected by spaces set before <li>
+    text = re.sub(r'([\r\n]<h4>)(.*)(</h4>[\r\n])(     <li>)', r'\1\2\3<ol>\n  <ol>\n    <ol class="bull">\n\4', text)
     text = re.sub(r'([\r\n]<p>)(.*)(</p>[\r\n])(     <li>)', r'\1\2\3<ol>\n  <ol>\n    <ol class="bull">\n\4', text)
     text = re.sub(r'([\r\n] <li>)(.*)(</li>[\r\n])(     <li>)', r'\1\2\3  <ol>\n    <ol class="bull">\n\4', text)
     text = re.sub(r'([\r\n]   <li>)(.*)(</li>[\r\n])(     <li>)', r'\1\2\3    <ol class="bull">\n\4', text)
@@ -222,7 +267,7 @@ def format_lists(text):
         text = re.sub(r'([\r\n]\s\s\s\s\s)(<li>)(.*)(</li>[\r\n])(\s[\s]?[\s]?<)', r'\1\2\3\4    </ol>\n\5', text)
 
         text = re.sub(r'([\r\n]\s\s\s\s)(</ol>)(.*)([\r\n])([\s]?<|\s\s<ol)', r'\1\2\3\4  </ol>\n\5', text)
-        text = re.sub(r'([\r\n]\s\s)(</ol>)(.*)([\r\n])(<h|<p)', r'\1\2\3\4</ol>\n\5', text)
+        text = re.sub(r'([\r\n]\s\s)(</ol>)(.*)([\r\n])(<h|<p|<t)', r'\1\2\3\4</ol>\n\5', text)
     return text
 
 
@@ -269,18 +314,21 @@ def format_bible_verses(text, title=None):
     return text
 
 
-def detect_language(text):
-    if pycld2 is None:
-        return None
-    raw_text = re.sub(r'[^a-zA-Z0-9\s]', r'', text)
-    reliable, index, top_3_choices = pycld2.detect(raw_text, bestEffort=False)
-    if not reliable:
-        reliable, index, top_3_choices = pycld2.detect(raw_text, bestEffort=True)
-    return top_3_choices[0][1]
+def format_ascii(text):
+    non_ascii = ['Ä', 'Ö', 'Ü', 'ä', 'ö', 'ü', 'ß', '–']
+    replace_with = ['&Auml;', '&Ouml;', '&Uuml;', '&auml;', '&ouml;', '&uuml;', '&szlig;', '&ndash;']
+    for x, y in zip(non_ascii, replace_with):
+        try:
+            text.replace(x, y)
+        except:
+            pass
+    return text
 
 
-def format_language(text):
-    if detect_language(text) == 'en':
+def format_language(text, lang=None):
+    if lang is None:
+        lang = detect_language(text)
+    if lang == 'en':
         text = text.replace('&ldquo;', '&rdquo;')
         text = text.replace('&lsquo;', '&rsquo;')
 
@@ -289,6 +337,12 @@ def format_language(text):
     return text
 
 
+def final_cut(text):
+    text = text.replace('<h2>', '<h1>')
+    text = text.replace('</h2>', '</h1>')
+    return text
+
+
 def assertion_test(text, text_orig, title):
     title_with_space = title + ' ' * (15 - len(re.sub(r'[^a-zA-Z0-9\s]', r'', title)))
 
@@ -298,12 +352,13 @@ def assertion_test(text, text_orig, title):
         if count_open != count_close:
             print('{} Unbalanced <{}>: {} <> {}'.format(title_with_space, item, count_open, count_close))
         if (item == 'ol') and (text_orig.count('<ol>') > 0):
-            print('{} {} <ol> in raw html detected'.format(' ' *len(title_with_space), text_orig.count("<ol>")))
+            text_sample = ' '.join(re.findall(r'<ol>(.*)</ol>', text_orig)[0].split()[:10])
+            print('{} {} <ol> in raw html detected: {}'.format(title_with_space, text_orig.count("<ol>"), text_sample))
 
     if text.count('. (Vers ') > 0:
         print('{} {} verse not correctly formatted'.format(title_with_space, text.count("(Vers ")))
     if text.count('style="padding-left') > 0:
-        print('{} Some bullets not correctly formatted'.format(title_with_space))
+        print('{} {} bullets not correctly formatted'.format(title_with_space), text.count('style="padding-left'))
 
     if False:  #check quotation marks later
         count_open, count_close = text.count('&bdquo;'), text.count('&ldquo;')
@@ -311,24 +366,6 @@ def assertion_test(text, text_orig, title):
             print('{} Unbalanced quotation marks: {} <> {}'.format(title_with_space, count_open, count_close))
 
 
-def spell_check(text):
-    from spellchecker import SpellChecker
-
-    text_raw = re.sub(r'<head>.*?</head>', r' ', text)
-    text_raw = re.sub(r'<.*?>|\(.*?\)', r' ', text_raw)
-    text_raw = re.sub(r'&bdquo;|&ldquo;|&copy;', r' ', text_raw)
-    text_raw = re.sub(r'[ ]?[iv]+\.', r' ', text_raw)
-    text_raw = re.sub(r',|\.|\?|\!|\;|\:|\(|\)', r'', text_raw)
-    text_raw = re.sub(r'  |   ', r' ', text_raw)
-
-    spell = SpellChecker(language='de', distance=1)
-    misspelled = spell.unknown(text_raw.split(' '))
-    for word in misspelled:
-        word_corrected = spell.correction(word)
-        if word != word_corrected:
-            print(word, spell.correction(word))
-
-
 """Markdown mappings"""
 
 def style_mappings_md(text):
@@ -342,4 +379,65 @@ def style_mappings_md(text):
     text = re.sub(r'([\r\n])([\r\n])', r'\1', text)
 
     text = re.sub(r'(\* \* \*[\r\n]„)(.*)(")', r'\1**\2** \3', text)
-    return text
\ No newline at end of file
+    return text
+
+
+"""Compare text with Schlachter bible text"""
+
+def load_bible():
+    file = os.path.join(os.path.dirname(__file__), 'Schlachter.txt')
+    bible = open(file, 'rb').read().decode('ISO-8859-1')
+    bible = re.sub(r'[^a-zA-Z0-9\s]', r'', bible)
+    bible = re.sub(r'([\r\n])', r' ', bible)
+    bible = re.sub(r' \d{1,3} ', r' ', bible)
+    bible = re.sub(r'  ', r' ', bible)
+    return bible
+
+
+def bible_check(text, title):
+    title_with_space = title + ' ' * (15 - len(re.sub(r'[^a-zA-Z0-9\s]', r'', title)))
+
+    bible = load_bible().lower()
+    verses = re.findall(r'<hr><p class="verse"> &bdquo;(.*)&ldquo;  <small>', text.lower())
+    verse_nums = re.findall(r'<hr><p class="verse"> &bdquo;.*&ldquo;  <small>\((.*)\)</small>', text.lower())
+
+    n = 0
+    for vers, vers_num in zip(verses, verse_nums):
+        vers = copy(re.sub(r'<br />|<small>|<small>|&bdquo;|&ldquo;', r'', vers))
+        vers = re.sub(r'[^a-zA-Z0-9\s]', r'', vers)
+        vers = re.sub(r'[\r\n]', r' ', vers)
+        vers = re.sub(r' \d{1,3} ', r' ', vers)
+        vers = re.sub(r'  ', r' ', vers)
+        valid = (vers[:30] in bible) or (vers[-30:] in bible)
+        n += valid
+
+        # for debugging:
+        #if not valid:
+        #    print(vers_num + '   ' + vers)
+
+    if len(verses) > 0:
+        perc_covered = int(n / len(verses) * 100)
+        if perc_covered < 60:
+            print('{} {}% of verses covered by Schlachter'.format(title_with_space, perc_covered))
+
+    return bible
+
+
+"""Spell Checker (currently not being used)"""
+
+def spell_check(text):
+    from spellchecker import SpellChecker
+
+    text_raw = re.sub(r'<head>.*?</head>', r' ', text)
+    text_raw = re.sub(r'<.*?>|\(.*?\)', r' ', text_raw)
+    text_raw = re.sub(r'&bdquo;|&ldquo;|&copy;', r' ', text_raw)
+    text_raw = re.sub(r'[ ]?[ivx]+\.', r' ', text_raw)
+    text_raw = re.sub(r',|\.|\?|\!|\;|\:|\(|\)', r'', text_raw)
+    text_raw = re.sub(r'  |   ', r' ', text_raw)
+
+    spell = SpellChecker(language='de', distance=1)
+    misspelled = spell.unknown(text_raw.split(' '))
+    for word in misspelled:
+        word_corrected = spell.correction(word)
+        if word != word_corrected:
+            print(word, spell.correction(word))
\ No newline at end of file
